""" File to contain string constants representing names used by Ringing Room to represent calls. """

LOOK_TO = "Look to"
GO = "Go"
BOB = "Bob"
SINGLE = "Single"
THATS_ALL = "That's all"
STAND = "Stand next"
